/*
   This file is part of GNOME Reversi, also known as Iagno.

   Copyright 2015, 2016, 2019 Arnaud Bonatti

   GNOME Reversi is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   GNOME Reversi is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNOME Reversi.  If not, see <https://www.gnu.org/licenses/>.
*/

@import url("base-window.css");

button.unfullscreen-button {
  margin:6px;
}

.extra-thin-window.thin-window button.history-button  { padding-left:12px; padding-right:12px; transition: padding 0 ease; }
.extra-thin-window.thin-window button.new-game-button { padding-left:12px; padding-right:12px; transition: padding 0 ease;
/* hack: fix the double spacing around the centerwidget box, on extra-thin window */
                                                        margin-right:-12px; }
/*\
* * board generics; TODO move in game-window.css
\*/

                               .game-box                { transition:padding 0.3s; padding:1.5rem; }
                  .thin-window .game-box                {                          padding:1.0rem; }
.extra-thin-window.thin-window .game-box,
                  .flat-window .game-box                {                          padding:0.5rem; }
.extra-flat-window.flat-window .game-box                {                          padding:0.4rem; }

/*\
* * start-game button; TODO move in game-window.css
\*/

                               button.start-game-button { margin-top:1.5rem; margin-bottom:0.5rem;
                                               transition:margin-top 0 ease, margin-bottom 0 ease; }
                  .flat-window button.start-game-button { margin-top:0.5rem;                       }
.extra-flat-window.flat-window button.start-game-button { margin-top:  0rem; margin-bottom:0.4rem; }

.extra-thin-window             button.start-game-button { margin-top:1.0rem; }

.extra-flat-window button.start-game-button {
  min-height:2rem;
}
                   button.start-game-button {
  min-height:3rem;
  min-width:11rem;

  transition:min-height          0.3s ease 0.01s;
}

/*\
* * labels' tweaks
\*/

/* "Default" theme is in italic */
popover > stack > box:nth-child(2) > box > box > box:first-child > modelbutton:first-child > box > label {
  font-style:italic;
}

/* make all widgets change font-size at same time */
label.bold-label,
button,
headerbar,
popover modelbutton {
  transition:all       0    ease,
             font-size 0.3s ease,
             padding   0.3s ease,
             margin    0.3s ease;
}

/*\
* * games buttons
\*/

button.radio-like-toggle#radio-one { background-image: url("resource://org/gnome/Reversi/images/reversi.png"); }
button.radio-like-toggle#radio-two { background-image: url("resource://org/gnome/Reversi/images/reverse.png"); }

/*\
* * the end
\*/
